/*
 *
 * File:         nbconv.c
 * Description:  Convert between N-byte and KS or TriGem code.
 * Author:       mleisher@nmsu.edu (Mark Leisher)
 * Created:      Wed Jan 30 15:06:23 1991
 * Modified:     
 *
 */

/*
 * Copyright (C) 1990 Mark Leisher.
 *
 * Author: Mark Leisher (mleisher@nmsu.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * A copy of the GNU General Public License can be obtained from this
 * program's author (send electronic mail to mleisher@nmsu.edu) or from
 * the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
 * 02139, USA.
 *
 */

#include <stdio.h>
#include "patchlevel.h"

#define NBCONV_VERSION 1

#define FALSE 0
#define TRUE  1

# ifndef N2KS_CONV_TABLE
# define N2KS_CONV_TABLE "./n2ks.table"
# endif /* N2KS_CONV_TABLE */
                            /* Default N-byte <-> KS conversion table.     */

#ifndef N2TG_CONV_TABLE
#define N2TG_CONV_TABLE "./n2trigem.table"
#endif /* N2TG_CONV_TABLE */
                            /* Default N-byte <-> TriGem conversion table. */


#define CTL_N        0x0e   /* N-byte string start indicator.              */
#define CTL_O        0x0f   /* N-byte string end indicator.                */
#define EOL          0x0a   /* Unix style ^J line feed.                    */

#define ONE      0          /* Initial(starting) state for N-byte parse.   */
#define TWO      1          /* Middle state for N-byte parse.              */
#define THREE    2          /* Final(ending) state for N-byte parse.       */

char *program;              /* Program name from the command line.         */

int from_nb = TRUE;         /* Converting from N-byte to KS or TriGem.     */
int do_ks = TRUE;           /* Convert between N-byte and KS (TRUE).       */
                            /* Convert between N-byte and TriGem (FALSE).  */
int doing_conv = FALSE;     /* Working on a converted string.              */
int ks_tbl_loaded = FALSE;  /* KS conversion table loaded.                 */
int tg_tbl_loaded = FALSE;  /* TriGem conversion table loaded.             */

/*
 * Hash table stuff.  Make a table of 4000 slots for 2390 items.
 */

#define M       4000
#define NORMAL  0
#define REVERSE 1

typedef struct {
    unsigned char *key;
    unsigned char *data;
} datanode;

typedef struct {
    datanode m1;
    datanode m2;
} hashnode;

static hashnode kstable[M];
static hashnode tgtable[M];

/*
 * Use Horner's method for evaluating a polynomial at a point.
 */
static int
hash(key)
unsigned char *key;
{
    int h;

    for (h = 0; *key != '\0'; key++)
      h = (64*h + *key) % M;

    return(h);
}

/*
 * A second hash function for double hashing with linear probing.
 * Uses the last 5 bits of the previously generated hash code.
 */
static int
hash2(h)
int h;
{
    return(32 - (h % 32));
}

void
hashinit()
{
    register int i;

    for (i = 0; i < M; i++) {
        kstable[i].m1.key = kstable[i].m1.data =
          kstable[i].m2.key = kstable[i].m2.data = NULL;
        tgtable[i].m1.key = tgtable[i].m1.data =
          tgtable[i].m2.key = tgtable[i].m2.data = NULL;
    }
}

/*
 * Insert with double hashing and linear probing.
 */
static void
hashinsert(key, data, keylength, direction)
unsigned char *key, *data;
int keylength, direction;
{
    register int h = hash(key);
    register int nh = hash2(h);
    register hashnode *table;

    table = (do_ks == TRUE) ? kstable : tgtable;

    if (direction == NORMAL) {
        while(table[h].m1.key != NULL &&
              strncmp((char *)table[h].m1.key, (char *)key, keylength))
          h = (h + nh) % M;
        table[h].m1.key = key;
        table[h].m1.data = data;
    } else {
        while(table[h].m2.key != NULL &&
              strncmp((char *)table[h].m2.key, (char *)key, keylength))
          h = (h + nh) % M;
        table[h].m2.key = key;
        table[h].m2.data = data;
    }
}

/*
 * Search for key.
 */
static unsigned char *
hashlookup(key, keylength, direction)
unsigned char *key;
int keylength, direction;
{
    register int h = hash(key);
    register int nh = hash2(h);
    register unsigned char *data = NULL;
    register hashnode *table;

    table = (do_ks == TRUE) ? kstable : tgtable;

    if (direction == NORMAL) {
        while(table[h].m1.key != NULL &&
              strncmp((char *)table[h].m1.key, (char *)key, keylength))
          h = (h + nh) % M;
        data = table[h].m1.data;
    } else {
        while(table[h].m2.key != NULL &&
              strncmp((char *)table[h].m2.key, (char *)key, keylength))
          h = (h + nh) % M;
        data = table[h].m2.data;
    }
    return(data);
}

static void
loadconvtable(filename)
char *filename;
{
    FILE *f;
    unsigned char buf[5];
    register unsigned char *nb, *code;
    register int c, i = 0, nsize = 0, cdsize = 2;

    if (!(f = fopen(filename, "r+b"))) {
	perror("fopen");
        fprintf(stderr, "%s: problem with conv table file %s.\n", program,
                filename);
        exit(-1);
    }

    hashinit();
    while(!feof(f)) {

        if (0 <= i && i < 40 && nsize != 1)
          nsize = 1;
        else if (40 <= i && i < 389 && nsize != 2)
          nsize = 2;
        else if (i >= 389 && nsize != 3)
          nsize = 3;

        if ((fread(buf, 1, (nsize+cdsize), f)) < (nsize+cdsize)) {
            fprintf(stderr, "%s: problem reading conv table file %s %d\n",
                    program, filename, i);
            fclose(f);
            exit(-1);
        }

        nb = (unsigned char *)malloc((nsize+1) * sizeof(unsigned char));
        code = (unsigned char *)malloc((cdsize+1) * sizeof(unsigned char));
        nb[nsize] = code[cdsize] = '\0';
        strncpy(nb, buf, nsize);
        strncpy(code, (buf+nsize), cdsize);

        /*
         * Insert by hashing on N-byte sequence.
         */
        hashinsert(nb, code, nsize, NORMAL);

        /*
         * Insert by hashing on KS or TriGem sequence.
         */
        hashinsert(code, nb, cdsize, REVERSE);

        i++;
        c = getc(f);
        while(!feof(f) && c < 0x20)
          c = getc(f);
        if (!feof(f))
          ungetc(c, f);
    }
    fclose(f);
}

static int
vowelcombine(c1, c2)
int c1, c2;
{
    if (from_nb) {
        switch(c1) {
          case 'l':
            switch(c2) {
              case 'b': return('K'); break;
              case 'c': return('H'); break;
              case '|': return('L'); break;
            }
            break;
          case 's':
            switch(c2) {
              case 'f': return('J'); break;
              case 'g': return('N'); break;
              case '|': return('B'); break;
            }
            break;
          case 'z':
            if (c2 == '|') return('M');
            break;
        }
    } else {
        switch(c1) {
          case 'K': return('b'); break;
          case 'H': return('c'); break;
          case 'L': return('|'); break;
          case 'J': return('f'); break;
          case 'N': return('g'); break;
          case 'B': return('|'); break;
          case 'M': return('|'); break;
        }
    }
    return(0);
}

static int
conscombine(c1, c2)
int c1, c2;
{
    if (from_nb) {
        switch(c1) {
          case 'A':
            if (c2 == 'U') return('S'); break;
          case 'D':
            switch(c2) {
              case 'X': return('J'); break;
              case '^': return('H'); break;
            }
            break;
          case 'I':
            switch(c2) {
              case 'A': return('G'); break;
              case 'Q': return('M'); break;
              case 'R': return('N'); break;
              case 'U': return('A'); break;
              case '\\': return('B'); break;
              case ']': return('V'); break;
              case '^': return('F'); break;
            }
            break;
          case 'R':
            if (c2 == 'U') return('C');
            break;
        }
    } else {
        switch(c1) {
          case 'S': return('U'); break;
          case 'J': return('X'); break;
          case 'H': return('^'); break;
          case 'G': return('A'); break;
          case 'M': return('Q'); break;
          case 'N': return('R'); break;
          case 'A': return('U'); break;
          case 'B': return('\\'); break;
          case 'V': return(']'); break;
          case 'F': return('^'); break;
          case 'C': return('U'); break;
        }
    }
    return(0);
}

static int
initcons(c)
int c;
{
    if (from_nb) {
        switch(c) {
          case 'A': return('r'); break;
          case 'B': return('R'); break;
          case 'D': return('s'); break;
          case 'G': return('e'); break;
          case 'H': return('E'); break;
          case 'I': return('f'); break;
          case 'Q': return('a'); break;
          case 'R': return('q'); break;
          case 'S': return('Q'); break;
          case 'U': return('t'); break;
          case 'V': return('T'); break;
          case 'W': return('d'); break;
          case 'X': return('w'); break;
          case 'Y': return('W'); break;
          case 'Z': return('c'); break;
          case '[': return('z'); break;
          case '\\': return('x'); break;
          case ']': return('v'); break;
          case '^': return('g'); break;
        }
    } else {
        switch(c) {
          case 'r': return('A'); break;
          case 'R': return('B'); break;
          case 's': return('D'); break;
          case 'e': return('G'); break;
          case 'E': return('H'); break;
          case 'f': return('I'); break;
          case 'a': return('Q'); break;
          case 'q': return('R'); break;
          case 'Q': return('S'); break;
          case 't': return('U'); break;
          case 'T': return('V'); break;
          case 'd': return('W'); break;
          case 'w': return('X'); break;
          case 'W': return('Y'); break;
          case 'c': return('Z'); break;
          case 'z': return('['); break;
          case 'x': return('\\'); break;
          case 'v': return(']'); break;
          case 'g': return('^'); break;
        }
    }
    return(0);
}

static int
vowel(c)
int c;
{
    if (from_nb) {
        switch (c) {
          case 'b': return('k'); break;
          case 'c': return('o'); break;
          case 'd': return('i'); break;
          case 'e': return('O'); break;
          case 'f': return('j'); break;
          case 'g': return('p'); break;
          case 'j': return('u'); break;
          case 'k': return('P'); break;
          case 'l': return('h'); break;
          case 'm': return('k'); break;
          case 'n': return('o'); break;
          case 'o': return('l'); break;
          case 'r': return('y'); break;
          case 's': return('n'); break;
          case 't': return('j'); break;
          case 'u': return('p'); break;
          case 'v': return('l'); break;
          case 'w': return('b'); break;
          case 'z': return('m'); break;
          case '{': return('M'); break;
          case '|': return('l'); break;
        }
    } else {
        switch (c) {
          case 'k': return('b'); break;
          case 'o': return('c'); break;
          case 'i': return('d'); break;
          case 'O': return('e'); break;
          case 'j': return('f'); break;
          case 'p': return('g'); break;
          case 'u': return('j'); break;
          case 'P': return('k'); break;
          case 'h': return('l'); break;
          case 'K': return('l'); break;
          case 'H': return('l'); break;
          case 'L': return('l'); break;
          case 'y': return('r'); break;
          case 'n': return('s'); break;
          case 'J': return('s'); break;
          case 'N': return('s'); break;
          case 'B': return('s'); break;
          case 'b': return('w'); break;
          case 'm': return('z'); break;
          case 'M': return('z'); break;
          case 'l': return('|'); break;
        }
    }
    return(0);
}

static int
finalcons(c)
int c;
{
    if (from_nb) {
        switch(c) {
          case 'A': return('r'); break;
          case 'B': return('R'); break;
          case 'C': return('S'); break;
          case 'D': return('s'); break;
          case 'E': return('J'); break;
          case 'F': return('H'); break;
          case 'G': return('e'); break;
          case 'I': return('f'); break;
          case 'J': return('G'); break;
          case 'K': return('M'); break;
          case 'L': return('N'); break;
          case 'M': return('A'); break;
          case 'N': return('B'); break;
          case 'O': return('V'); break;
          case 'P': return('F'); break;
          case 'Q': return('a'); break;
          case 'R': return('q'); break;
          case 'S': return('t'); break;
          case 'U': return('t'); break;
          case 'V': return('T'); break;
          case 'W': return('d'); break;
          case 'X': return('w'); break;
          case 'Z': return('c'); break;
          case '[': return('z'); break;
          case '\\': return('x'); break;
          case ']': return('v'); break;
          case '^': return('g'); break;
        }
    } else {
        switch(c) {
          case 'r': return('A'); break;
          case 'R': return('B'); break;
          case 'S': return('A'); break;
          case 's': return('D'); break;
          case 'J': return('D'); break;
          case 'H': return('D'); break;
          case 'e': return('G'); break;
          case 'f': return('I'); break;
          case 'G': return('I'); break;
          case 'M': return('I'); break;
          case 'N': return('I'); break;
          case 'A': return('I'); break;
          case 'B': return('I'); break;
          case 'V': return('I'); break;
          case 'F': return('I'); break;
          case 'a': return('Q'); break;
          case 'q': return('R'); break;
          case 'C': return('R'); break;
          case 't': return('U'); break;
          case 'T': return('V'); break;
          case 'd': return('W'); break;
          case 'w': return('X'); break;
          case 'c': return('Z'); break;
          case 'z': return('['); break;
          case 'x': return('\\'); break;
          case 'v': return(']'); break;
          case 'g': return('^'); break;
        }
    }
    return(0);
}

static void
map_from_nb(c1, c2, c3)
int c1, c2, c3;
{
    unsigned char key[4];
    unsigned char *code;
    int i = 0;

    if (!c1 && !c2 && !c3)
      return;
    if (c1) key[i++] = c1;
    if (c2) key[i++] = c2;
    if (c3) key[i++] = c3;

    key[i] = '\0';
    /*
     * Map unknown sequences to the space character because they probably
     * don't have equivalents in the KSC5601 code set.
     * The code position for a space in KS code is 0xa4a0.
     */
    if (!(code = hashlookup(key, i, NORMAL)))
      printf("%c%c", 0xa4, 0xa0);
    else
      printf("%s", code);
}

static void
map_to_nb(ks)
unsigned char *ks;
{
    unsigned char *ncode = hashlookup(ks, 2, REVERSE);
    int i, len, res;

    if (!ncode)
      return;

    len = strlen(ncode);
    for (i = ONE; i < len; i++) {
        switch(i) {
          case ONE:
            if ((res = initcons(ncode[i])))
              putchar(res);
            else if ((res = vowel(ncode[i])))
              putchar(res);
            else if ((res = finalcons(ncode[i])))
              putchar(res);
            else
              fprintf(stderr, "%s: illegal char %c[%d]\n", program,
                      ncode[i], ncode[i]);
            break;
          case TWO:
            if ((res = vowel(ncode[i]))) {
              putchar(res);
              if ((res = vowelcombine(ncode[i], 0)))
                putchar(res);
            } else if ((res = vowelcombine(ncode[i], 0)))
              putchar(res);
            else
              fprintf(stderr, "%s: illegal char %c[%d]\n", program,
                      ncode[i], ncode[i]);
            break;
          case THREE:
            if ((res = finalcons(ncode[i]))) {
                putchar(res);
                if ((res = conscombine(ncode[i], 0)))
                  putchar(res);
            } else if ((res = conscombine(ncode[i], 0)))
              putchar(res);
            else
              fprintf(stderr, "%s: illegal char %c[%d]\n", program,
                      ncode[i], ncode[i]);
            break;
        }
    }
}

static void
conv_from_nb(f)
FILE *f;
{
    int c, c1, c2, res;
    int cho_sung = 0, jung_sung = 0, jong_sung = 0;
    int continue_exit, state = 0;


    c = getc(f);
    while(!feof(f)) {
        continue_exit = FALSE;
        if (c == CTL_N) {
            doing_conv = TRUE;
            continue_exit = TRUE;
            state = ONE;

            /*
             * Only load the conversion table if something to convert.
             */
            if (do_ks && !ks_tbl_loaded) {
                loadconvtable(N2KS_CONV_TABLE);
                ks_tbl_loaded = TRUE;
            } else if (!do_ks && !tg_tbl_loaded) {
                loadconvtable(N2TG_CONV_TABLE);
                tg_tbl_loaded = TRUE;
            }
        } else if (c == CTL_O) {
            map_from_nb(cho_sung, jung_sung, jong_sung);
            cho_sung = jung_sung = jong_sung = 0;
            doing_conv = FALSE;
            continue_exit = TRUE;
            state = ONE;
        }
        if (continue_exit) {
            c = getc(f);
            continue;
        }
        if (doing_conv) {
            switch(state) {
              case ONE:
                if ((cho_sung = initcons(c)))
                    state = TWO;
                else if ((jung_sung = vowel(c))) {
                    cho_sung = jong_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                } else if ((jong_sung = finalcons(c))) {
                    cho_sung = jung_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                }
                break;
              case TWO:
                c1 = getc(f);
                if ((jung_sung = vowelcombine(c, c1))) {
                    state = THREE;
                    c1 = 0;
                } else if ((jung_sung = vowel(c))) {
                    state = THREE;
                } else if ((res = initcons(c))) {
                    jung_sung = jong_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    cho_sung = res;
                } else if ((res = finalcons(c))) {
                    jung_sung = jong_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    cho_sung = 0;
                    jong_sung = res;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    jong_sung = 0;
                    state = ONE;
                }
                if (!feof(f) && c1)
                  ungetc(c1, f);
                c1 = 0;
                break;
              case THREE:
                if ((c1 = getc(f)) != EOF)
                  c2 = getc(f);
                else
                  c2 = 0;
                if ((res = initcons(c)) && vowel(c1)) {
                    jong_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    cho_sung = res;
                    state = TWO;
                    if (c2 && c2 != EOF)
                      ungetc(c2, f);
                    if (c1 && c1 != EOF)
                      ungetc(c1, f);
                    c1 = c2 = 0;
                } else if ((jong_sung = conscombine(c, c1)) && !vowel(c2)) {
                    if (c2 && c2 != EOF)
                      ungetc(c2, f);
                    c1 = c2 = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    state = ONE;
                    cho_sung = jung_sung = jong_sung = 0;
                } else if ((jong_sung = finalcons(c))) {
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    cho_sung = jung_sung = jong_sung = 0;
                    state = ONE;
                } else if ((res = initcons(c))) {
                    jong_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    cho_sung = res;
                    state = TWO;
                } else if ((res = vowel(c))) {
                    jong_sung = 0;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    cho_sung = 0;
                    jung_sung = res;
                    map_from_nb(cho_sung, jung_sung, jong_sung);
                    jung_sung = 0;
                    state = ONE;
                }
                if (c2 && c2 != EOF)
                  ungetc(c2, f);
                if (c1 && c1 != EOF)
                  ungetc(c1, f);
                break;
            }
        } else
          putchar(c);
        c = getc(f);
    }
}

static void
conv_to_nb(f)
FILE *f;
{
    int c, c1;
    unsigned char code[3];

    code[2] = '\0';
    c = getc(f);
    while(!feof(f)) {
        if (c & 0x80) {
            if ((c1 = getc(f)) == EOF) {
                if (doing_conv) {
                    putchar(CTL_O);
                    doing_conv = FALSE;
                }
                putchar(c);
                continue;
            }
            if (!doing_conv) {
                putchar(CTL_N);
                doing_conv = TRUE;

                /*
                 * Only load the conversion table if something to convert.
                 */
                if (do_ks && !ks_tbl_loaded) {
                    loadconvtable(N2KS_CONV_TABLE);
                    ks_tbl_loaded = TRUE;
                } else if (!do_ks && !tg_tbl_loaded) {
                    loadconvtable(N2TG_CONV_TABLE);
                    tg_tbl_loaded = TRUE;
                }
            }
            code[0] = c;
            code[1] = c1;
            map_to_nb(code);
        } else if (doing_conv) {
            putchar(CTL_O);
            doing_conv = FALSE;
            putchar(c);
        } else
          putchar(c);
        c = getc(f);
    }
}

static void
version()
{
    fprintf(stderr, "%s version %d.%d\n", program, NBCONV_VERSION, PATCHLEVEL);
    exit(1);
}

static void
help()
{
    fprintf(stderr, "usage:  %s [-nk] [-kn] [-nt] [-tn] [-h] [-v] [-]\n",
            program);
    fprintf(stderr, "\t-nk\t convert from N-byte to KS (default)\n");
    fprintf(stderr, "\t-kn\t convert from KS to N-byte\n");
    fprintf(stderr, "\t-nt\t convert from N-byte to TriGem\n");
    fprintf(stderr, "\t-tn\t convert from TriGem to N-byte\n");
    fprintf(stderr, "\t-h\t this message\n");
    fprintf(stderr, "\t-v\t print the program version\n");
    fprintf(stderr, "\t-\t read from stdin\n");
    exit(1);
}

main(argc, argv)
int argc;
char **argv;
{
    FILE *in = NULL;

    program = argv[0];

    *argv++;
    argc--;
    while(argc) {
        if (argv[0][0] == '-') {
            switch(argv[0][1]) {
              case 'k': case 'K':
                if (argv[0][2] != 'n' && argv[0][2] != 'N') {
                    fprintf(stderr, "%s: unknown argument %s\n\n", program,
                            argv[0]);
                    help();
                }                    
                do_ks = TRUE;
                from_nb = FALSE;
                break;
              case 't': case 'T':
                if (argv[0][2] != 'n' && argv[0][2] != 'N') {
                    fprintf(stderr, "%s: unknown argument %s\n\n", program,
                            argv[0]);
                    help();
                }                    
                do_ks = FALSE;
                from_nb = FALSE;
                break;
              case 'n': case 'N':
                from_nb = TRUE;
                switch(argv[0][2]) {
                  case 'k': case 'K':
                    do_ks = TRUE;
                    break;
                  case 't': case 'T':
                    do_ks = FALSE;
                    break;
                  default:
                    fprintf(stderr, "%s: unknown argument %s\n\n", program,
                            argv[0]);
                    help();
                    break;
                }
                break;
              case 'h': case 'H':
                help();
                break;
              case 'v': case 'V':
                version();
                break;
              default:
                in = stdin;
                break;
            }
            if (in != NULL) {
                if (from_nb)
                  conv_from_nb(in);
                else
                  conv_to_nb(in);
                in = NULL;
            }
        } else {
            if (!(in = fopen(argv[0], "r"))) {
                fprintf(stderr, "%s: problem with input file %s\n", program,
                        argv[0]);
                exit(-1);
            }
            if (from_nb)
              conv_from_nb(in);
            else
              conv_to_nb(in);
            if (in != stdin)
              fclose(in);
            in = NULL;
        }
        *argv++;
        argc--;
    }
    exit(0);
}
